# coding: utf-8
from functools import singledispatch
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse


def add_query_params(url, **kwargs):
    """
    Adds extra querystring parameters to given URL.
    """
    parsed = list(urlparse(url))
    get_args = parse_qsl(parsed[4]) + [
        (sub_key, sub_value)
        for key, value in kwargs.items()
        for sub_key, sub_value in expand(value, key)
        if sub_value is not None
    ]
    parsed[4] = urlencode(get_args)
    return urlunparse(parsed)


@singledispatch
def expand(value, key):
    yield key, value


@expand.register(dict)
def expand_dict(value, key):
    for inner_key, inner_value in value.items():
        yield '%s[%s]' % (key, inner_key), inner_value


@expand.register(list)
@expand.register(tuple)
def expand_lists(value, key):
    for inner_value in value:
        yield key, inner_value



# coding: utf-8
from .core import Model, ModelManager


class Advertiser(Model):
    """
    An Advertiser account.
    """

    def update(self):
        return self._manager.update(**self.instance_data)

    def block(self, reason=''):
        return self._manager.block(self['id'], reason)


class AdvertiserManager(ModelManager):
    model = 'Advertiser'

    def create(self, **kwargs):
        return self._call('create', data=kwargs)

    def update(self, id, **kwargs):
        return self._call('update', id=id, data=kwargs)

    def find_by_id(self, id):
        return self._call('findById', id=id)

    def block(self, id, reason=''):
        return self._call('block', id=id, reason=reason)



# coding: utf-8
import requests

from .exceptions import HasOffersException
from .logging import get_logger
from .models.advertiser import Advertiser, AdvertiserManager
from .models.goal import Goal, GoalManager
from .models.offer import Offer, OfferManager
from .utils import add_query_params


class HasOffersAPI:
    """
    Client to communicate with HasOffers API.
    """
    managers = {
        'advertisers': AdvertiserManager,
        'goals': GoalManager,
        'offers': OfferManager,
    }

    def __init__(self, endpoint=None, network_token=None, network_id=None, verbosity=0):
        self.endpoint = endpoint
        self.network_token = network_token
        self.network_id = network_id
        self.logger = get_logger(verbosity)
        for name, manager in self.managers.items():
            setattr(self, name, manager(self))

    def __str__(self):
        return '%s: %s / %s' % (self.__class__.__name__, self.network_token, self.network_id)

    def __repr__(self):
        return '<%s>' % self

    @property
    def session(self):
        if not hasattr(self, '_session'):
            self._session = requests.Session()
        return self._session

    def _call(self, target, method, **kwargs):
        """
        Low-level call to HasOffers API.
        """
        params = {
            'NetworkToken': self.network_token,
            'NetworkId': self.network_id,
            'Target': target,
            'Method': method
        }
        params.update(**kwargs)
        url = add_query_params(self.endpoint, **params)
        response = self.session.get(url, verify=False)
        response.raise_for_status()
        content = response.json()
        self.logger.debug('Response: %s', content)
        return self.handle_response(content)

    def handle_response(self, content):
        """
        Parses response, checks it.
        """
        response = content['response']

        errors = response.get('errors')
        if errors:
            raise HasOffersException(errors)

        data = response.get('data')

        if isinstance(data, bool) or data is None:
            return data
        if 'Advertiser' in data:
            return Advertiser(manager=self.advertisers, **data['Advertiser'])
        elif 'Offer' in data:
            return Offer(manager=self.offers, **data['Offer'])
        elif 'Goal' in data:
            return Goal(manager=self.goals, **data['Goal'])



# coding: utf-8
from .core import Model, ModelManager


class Offer(Model):
    """
    An Offer.
    """
    non_model_fields = ('dne_third_party_list', )

    def update(self):
        return self._manager.update(**self.instance_data)

    def add_target_country(self, country_code):
        return self._manager.add_target_country(self['id'], country_code)

    def add_category(self, category_id):
        return self._manager.add_category(self['id'], category_id)


class OfferManager(ModelManager):
    model = 'Offer'

    def create(self, **kwargs):
        return self._call('create', data=kwargs)

    def update(self, id, **kwargs):
        return self._call('update', id=id, data=kwargs)

    def find_by_id(self, id):
        return self._call('findById', id=id)

    def add_target_country(self, id, country_code):
        return self._call('addTargetCountry', id=id, country_code=country_code)

    def add_category(self, id, category_id):
        return self._call('addCategory', id=id, category_id=category_id)



# coding: utf-8


class Model:
    """
    Abstract model for HasOffers entity.
    """
    non_model_fields = ()

    def __init__(self, manager, **kwargs):
        self._manager = manager
        self._data = kwargs

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self['id'] == other['id']

    @property
    def instance_data(self):
        """
        All data except for private and non model fields.
        """
        return {
            key: value for key, value in self._data.items()
            if not key.startswith('_') and key not in self.non_model_fields
        }


class ModelManager:
    """
    Proxy for API methods with predefined model.
    Used to structure API. Every manager works only with single model methods.
    """
    model = None

    def __init__(self, api):
        self.api = api

    def _call(self, method, **kwargs):
        return self.api._call(self.model, method, **kwargs)
